from kivy.base import EventLoop

from ..pallet import Pallet
from .util import ELiDEAppTest, idle_until


class TestSpriteBuilder(ELiDEAppTest):
	def test_build_pawn(self):
		app = self.app
		self.Window.add_widget(app.build())
		app.manager.current = "pawncfg"
		idle_until(
			lambda: "dialog" in app.pawncfg.ids,
			100,
			"Never made dialog for pawncfg",
		)
		pawn_cfg_dialog = app.pawncfg.ids.dialog
		idle_until(
			lambda: "builder" in pawn_cfg_dialog.ids,
			100,
			"Never made pawn builder",
		)
		builder = pawn_cfg_dialog.ids.builder
		idle_until(lambda: builder.labels, 100, "Never got any builder labels")
		idle_until(
			lambda: builder.pallets, 100, "Never got any builder pallets"
		)
		idle_until(
			lambda: len(builder.labels) == len(builder.pallets),
			100,
			"Never updated pawn builder",
		)
		palbox = builder._palbox
		for child in palbox.children:
			if not isinstance(child, Pallet):
				continue
			idle_until(
				lambda: child.swatches,
				100,
				"Never got swatches for " + child.filename,
			)
			if "draconian_m" in child.swatches:
				child.swatches["draconian_m"].state = "down"
				idle_until(
					lambda: child.swatches["draconian_m"] in child.selection,
					100,
					"Selection never updated",
				)
			if "robe_red" in child.swatches:
				child.swatches["robe_red"].state = "down"
				idle_until(
					lambda: child.swatches["robe_red"] in child.selection,
					100,
					"Selection never updated",
				)
		idle_until(
			lambda: pawn_cfg_dialog.ids.selector.imgpaths,
			100,
			"Never got imgpaths",
		)
		pawn_cfg_dialog.pressed()
		idle_until(
			lambda: pawn_cfg_dialog.imgpaths, 100, "Never propagated imgpaths"
		)
		assert pawn_cfg_dialog.imgpaths == [
			"atlas://base.atlas/draconian_m",
			"atlas://body.atlas/robe_red",
		]



import networkx as nx
from kivy.base import EventLoop
from kivy.tests.common import GraphicUnitTest

from elide.grid.board import GridBoard, GridBoardView
from lisien import Engine
from lisien.facade import CharacterFacade

from .util import ELiDEAppTest, all_pawns_placed, all_spots_placed, idle_until


class GridBoardTest(GraphicUnitTest):
	def test_layout_grid(self):
		spots_wide = 3
		spots_tall = 3
		spot_width = 32
		spot_height = 32
		graph = nx.grid_2d_graph(spots_wide, spots_tall)
		char = CharacterFacade(graph)
		char.place[1, 1].add_thing("something")
		otherthing = char.place[2, 2].new_thing("otherthing")
		assert len(char.thing) == 2
		board = GridBoard(character=char)
		self.Window.add_widget(GridBoardView(board=board))
		while not (
			all_spots_placed(board, char) and all_pawns_placed(board, char)
		):
			EventLoop.idle()
		otherthing["location"] = (0, 0)
		board.spot_plane.data = list(map(board.make_spot, char.place.values()))
		board.spot_plane.redraw()
		board.pawn_plane.data = list(map(board.make_pawn, char.thing.values()))

		def arranged():
			for x in range(spots_wide):
				for y in range(spots_tall):
					spot = board.spot[x, y]
					if spot.x != x * spot_width or spot.y != y * spot_height:
						return False
			return True

		idle_until(arranged, 100)
		idle_until(lambda: board.pawn_plane._stack_index)
		this = board.pawn["something"]
		that = board.pawn["otherthing"]
		print(this.pos, board.spot[1, 1].pos)
		idle_until(lambda: this.pos == board.spot[1, 1].pos)
		idle_until(lambda: that.pos == board.spot[0, 0].pos)
		assert this.x == board.spot[1, 1].x
		assert this.y == board.spot[1, 1].y
		assert that.x == board.spot[0, 0].x
		assert that.y == board.spot[0, 0].y


class SwitchGridTest(ELiDEAppTest):
	def test_character_switch_grid(self):
		with Engine(self.prefix) as eng:
			eng.add_character("physical", nx.grid_2d_graph(10, 1))
			eng.add_character("tall", nx.grid_2d_graph(1, 10))
		app = self.app
		self.Window.add_widget(app.build())
		idle_until(
			lambda: hasattr(app, "mainscreen")
			and app.mainscreen.mainview
			and app.mainscreen.statpanel
			and hasattr(app.mainscreen, "gridview")
		)
		app.mainscreen.charmenu.toggle_gridview()
		idle_until(
			lambda: app.mainscreen.gridview in app.mainscreen.mainview.children
		)
		idle_until(lambda: app.mainscreen.gridview.board.children)
		assert len(app.mainscreen.gridview.board.spot) == 10
		assert all(
			spot.y == 0 for spot in app.mainscreen.gridview.board.spot.values()
		)
		idle_until(
			lambda: not all(
				spot.x == 0
				for spot in app.mainscreen.gridview.board.spot.values()
			),
			100,
		)
		app.character_name = "tall"
		idle_until(
			lambda: all(
				spot.x == 0
				for spot in app.mainscreen.gridview.board.spot.values()
			),
			1000,
			"Never got the new board",
		)
		idle_until(
			lambda: not all(
				spot.y == 0
				for spot in app.mainscreen.gridview.board.spot.values()
			),
			1000,
			"New board arranged weird",
		)



from lisien import Engine
from lisien.examples import sickle

from .util import ELiDEAppTest, idle_until


class PythonEditorTest(ELiDEAppTest):
	def setUp(self):
		super().setUp()
		with Engine(self.prefix) as eng:
			sickle.install(eng)

	def _get_actions_box(self):
		app = self.app
		idle_until(
			lambda: hasattr(app, "mainscreen")
			and app.mainscreen.mainview
			and app.mainscreen.statpanel
			and hasattr(app.mainscreen, "gridview")
		)
		app.funcs.toggle()
		idle_until(
			lambda: "actions" in app.funcs.ids, 100, "Never got actions box"
		)
		actions_box = app.funcs.ids.actions
		idle_until(lambda: actions_box.editor, 100, "Never got FuncEditor")
		idle_until(lambda: actions_box.storelist, 100, "Never got StoreList")
		idle_until(
			lambda: actions_box.storelist.data, 100, "Never got StoreList data"
		)
		return actions_box


class TestShowCode(PythonEditorTest):
	def test_show_code(self):
		app = self.app
		self.Window.add_widget(app.build())
		actions_box = self._get_actions_box()
		last = actions_box.storelist.data[-1]["name"]
		actions_box.storelist.selection_name = last
		idle_until(
			lambda: "funname" in actions_box.editor.ids,
			100,
			"Never got function input widget",
		)
		idle_until(
			lambda: actions_box.editor.ids.funname.hint_text,
			100,
			"Never got function name",
		)
		idle_until(
			lambda: "code" in actions_box.editor.ids,
			100,
			"Never got code editor widget",
		)
		idle_until(
			lambda: actions_box.editor.ids.code.text,
			100,
			"Never got source code",
		)


class TestCreateAction(PythonEditorTest):
	def test_create_action(self):
		app = self.app
		self.Window.add_widget(app.build())
		actions_box = self._get_actions_box()
		actions_box.editor.ids.funname.text = "new_func"
		actions_box.editor.ids.code.text = 'return "Hello, world!"'
		app.stop()
		with Engine(self.prefix) as eng:
			assert hasattr(eng.action, "new_func")



import shutil
import sys
from functools import partial
from tempfile import mkdtemp

from blinker import Signal
from kivy.base import EventLoop
from kivy.config import ConfigParser
from kivy.input.motionevent import MotionEvent
from kivy.tests.common import GraphicUnitTest

from elide.app import ElideApp


def all_spots_placed(board, char=None):
	if char is None:
		char = board.character
	for place in char.place:
		if place not in board.spot:
			return False
	return True


def all_pawns_placed(board, char=None):
	if char is None:
		char = board.character
	for thing in char.thing:
		if thing not in board.pawn:
			return False
	return True


def all_arrows_placed(board, char=None):
	if char is None:
		char = board.character
	for orig, dests in char.portal.items():
		if orig not in board.arrow:
			return False
		arrows = board.arrow[orig]
		for dest in dests:
			if dest not in arrows:
				return False
	return True


def board_is_arranged(board, char=None):
	if char is None:
		char = board.character
	return (
		all_spots_placed(board, char)
		and all_pawns_placed(board, char)
		and all_arrows_placed(board, char)
	)


def idle_until(condition=None, timeout=100, message="Timed out"):
	"""Advance frames until ``condition()`` is true

	With integer ``timeout``, give up after that many frames,
	raising ``TimeoutError``. You can customize its ``message``.

	"""
	if not (timeout or condition):
		raise ValueError("Need timeout or condition")
	if condition is None:
		return partial(idle_until, timeout=timeout, message=message)
	if timeout is None:
		while not condition():
			EventLoop.idle()
		return
	for _ in range(timeout):
		if condition():
			return
		EventLoop.idle()
	raise TimeoutError(message)


class MockTouch(MotionEvent):
	def depack(self, args):
		self.is_touch = True
		self.sx = args["sx"]
		self.sy = args["sy"]
		super().depack(args)


class ListenableDict(dict, Signal):
	def __init__(self):
		Signal.__init__(self)


class MockTime(Signal):
	pass


class MockEngine(Signal):
	eternal = ListenableDict()
	universal = ListenableDict()
	character = ListenableDict()
	string = ListenableDict()
	time = MockTime()
	closed = False

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.turn = self.initial_turn = self.final_turn = 0
		self._ready = True

	def __setattr__(self, key, value):
		if not hasattr(self, "_ready"):
			super().__setattr__(key, value)
			return
		self.send(self, key=key, value=value)
		super().__setattr__(key, value)

	def branch_start_turn(self, branch: str | None = None) -> int:
		return self.initial_turn

	def branch_end_turn(self, branch=None):
		return self.final_turn

	def next_turn(self, *args, **kwargs):
		self.turn += 1
		self.final_turn = self.turn
		kwargs["cb"]("next_turn", "trunk", self.turn, 0, ([], {}))

	def handle(self, *args, **kwargs):
		return {"a": "b"}

	def commit(self):
		pass


class ELiDEAppTest(GraphicUnitTest):
	def __init__(self, methodName="runTest"):
		super().__init__(methodName)
		self.prefix = mkdtemp()
		self.addCleanup(self.cleanup)

	def cleanup(self):
		shutil.rmtree(self.prefix)

	def setUp(self):
		super(ELiDEAppTest, self).setUp()
		self.old_argv = sys.argv.copy()
		sys.argv = ["python", "-m", "elide", self.prefix]
		self.app = ElideApp()
		self.app.config = ConfigParser(None)
		self.app.build_config(self.app.config)

	def tearDown(self, fake=False):
		EventLoop.idle()
		super().tearDown(fake=fake)
		self.app.stop()
		sys.argv = self.old_argv



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from kivy.lang import Builder
from kivy.uix.behaviors import FocusBehavior
from kivy.uix.recycleboxlayout import RecycleBoxLayout
from kivy.uix.recycleview.layout import LayoutSelectionBehavior

loaded_kv = set()


def load_string_once(kv: str) -> None:
	if kv in loaded_kv:
		return
	Builder.load_string(kv)
	loaded_kv.add(kv)


class SelectableRecycleBoxLayout(
	FocusBehavior, LayoutSelectionBehavior, RecycleBoxLayout
):
	pass


def dummynum(character, name):
	"""Count how many nodes there already are in the character whose name
	starts the same.

	"""
	num = 0
	for nodename in character.node:
		nodename = str(nodename)
		if nodename[: len(name)] != name:
			continue
		try:
			nodenum = int(nodename.lstrip(name))
		except ValueError:
			continue
		num = max((nodenum, num))
	return num



from kivy.tests.common import UnitTestTouch

from lisien import Engine
from lisien.examples import polygons

from .util import ELiDEAppTest, idle_until


class CharacterSwitcherTest(ELiDEAppTest):
	def setUp(self):
		super(CharacterSwitcherTest, self).setUp()
		with Engine(self.prefix) as eng:
			polygons.install(eng)

	def test_character_switcher(self):
		app = self.app
		self.Window.add_widget(app.build())
		idle_until(
			lambda: app.manager.current == "main",
			100,
			"Never switched to main",
		)
		idle_until(
			lambda: app.mainscreen.boardview, 100, "never got boardview"
		)
		idle_until(
			lambda: app.mainscreen.boardview.board.spot, 100, "never got spots"
		)
		physspots = len(app.mainscreen.boardview.board.spot)
		app.mainscreen.charmenu.charmenu.toggle_chars_screen()
		idle_until(
			lambda: app.manager.current == "chars",
			100,
			"Never switched to chars",
		)
		boxl = app.chars.ids.charsview.ids.boxl
		idle_until(
			lambda: len(boxl.children) == 3,
			100,
			"Didn't get all three characters",
		)
		for charb in boxl.children:
			if charb.text == "triangle":
				touch = UnitTestTouch(*charb.center)
				touch.pos = charb.center
				assert charb.dispatch("on_touch_down", touch)
				self.advance_frames(5)
				charb.dispatch("on_touch_up", touch)
				idle_until(
					lambda: charb.state == "down",
					10,
					"Button press did not work",
				)
				break
		else:
			assert False, 'No button for "triangle" character'
		idle_until(
			lambda: app.chars.ids.charsview.character_name == "triangle",
			100,
			"Never propagated character_name",
		)
		app.chars.toggle()
		idle_until(
			lambda: app.manager.current == "main",
			100,
			"Didn't switch back to main",
		)
		idle_until(
			lambda: not app.mainscreen.boardview.board.spot,
			100,
			"Didn't clear out spots, {} left".format(
				len(app.mainscreen.boardview.board.spot)
			),
		)
		app.mainscreen.charmenu.charmenu.toggle_chars_screen()
		idle_until(
			lambda: app.manager.current == "chars",
			100,
			"Never switched to chars",
		)
		for charb in boxl.children:
			if charb.text == "physical":
				touch = UnitTestTouch(*charb.center)
				touch.pos = charb.center
				assert charb.dispatch("on_touch_down", touch)
				self.advance_frames(5)
				charb.dispatch("on_touch_up", touch)
				idle_until(
					lambda: charb.state == "down",
					10,
					"Button press did not work",
				)
				break
		else:
			assert False, 'No button for "physical" character'
		idle_until(
			lambda: app.chars.ids.charsview.character_name == "physical",
			100,
			"Never propagated character_name",
		)
		app.chars.toggle()
		idle_until(
			lambda: len(app.mainscreen.boardview.board.spot) == physspots,
			100,
			"Never got physical back",
		)



from functools import partial
from types import SimpleNamespace

from kivy.base import EventLoop
from kivy.tests.common import UnitTestTouch
from kivy.uix.screenmanager import ScreenManager

from elide.graph.board import GraphBoard
from elide.grid.board import GridBoard
from elide.menu import DirPicker
from elide.screen import MainScreen
from elide.spritebuilder import PawnConfigScreen, SpotConfigScreen
from elide.statcfg import StatScreen
from lisien import Engine
from lisien.facade import CharacterFacade

from .util import ELiDEAppTest, ListenableDict, MockEngine, idle_until


class MockStore:
	def save(self, *args):
		pass


class ScreenTest(ELiDEAppTest):
	def test_advance_time(self):
		app = self.app
		app.mainmenu = DirPicker()
		app.spotcfg = SpotConfigScreen()
		app.pawncfg = PawnConfigScreen()
		app.statcfg = StatScreen()
		char = CharacterFacade("physical")
		app.character = char
		app.engine = MockEngine()
		app.strings = MockStore()
		app.funcs = MockStore()
		char.character = SimpleNamespace(engine=app.engine)
		app.engine.character["physical"] = char
		entity = ListenableDict()
		entity.engine = app.engine
		entity.name = "name"
		app.selected_proxy = app.proxy = app.statcfg.proxy = entity
		screen = MainScreen(
			graphboards={"physical": GraphBoard(character=char)},
			gridboards={"physical": GridBoard(character=char)},
		)
		self.Window.add_widget(screen)
		idle_until(
			lambda: "timepanel" in screen.ids, 100, "timepanel never got id"
		)
		timepanel = screen.ids["timepanel"]
		idle_until(
			lambda: timepanel.size != [100, 100],
			100,
			"timepanel never resized",
		)
		turnfield = timepanel.ids["turnfield"]
		turn_before = int(turnfield.hint_text)
		stepbut = timepanel.ids["stepbut"]
		motion = UnitTestTouch(*stepbut.center)
		motion.touch_down()
		motion.touch_up()
		EventLoop.idle()
		assert int(turnfield.hint_text) == turn_before + 1

	def test_play(self):
		app = self.app
		app.spotcfg = SpotConfigScreen()
		app.pawncfg = PawnConfigScreen()
		app.statcfg = StatScreen()
		app.mainmenu = DirPicker()
		app.strings = MockStore()
		app.funcs = MockStore()
		char = CharacterFacade("foo")
		app.character = char
		app.engine = MockEngine()
		app.manager = ScreenManager()
		char.character = SimpleNamespace(engine=app.engine)
		app.engine.character["foo"] = char
		entity = ListenableDict()
		entity.engine = app.engine
		entity.name = "name"
		app.selected_proxy = app.proxy = app.statcfg.proxy = entity
		screen = MainScreen(
			graphboards={"foo": GraphBoard(character=char)},
			gridboards={"foo": GridBoard(character=char)},
			play_speed=1.0,
		)
		app.manager.add_widget(screen)
		self.Window.add_widget(app.manager)
		idle_until(
			lambda: "timepanel" in screen.ids, 100, "timepanel never got id"
		)
		timepanel = screen.ids["timepanel"]
		idle_until(
			lambda: timepanel.size != [100, 100],
			100,
			"timepanel never resized",
		)
		turnfield = timepanel.ids["turnfield"]
		playbut = screen.playbut = timepanel.ids["playbut"]
		motion = UnitTestTouch(*playbut.center)
		motion.touch_down()
		motion.touch_up()
		idle_until(
			lambda: int(turnfield.hint_text) == 3,
			400,
			"Time didn't advance fast enough",
		)
		playbut.state = "normal"

	def test_update(self):
		def almost(a, b):
			if isinstance(a, tuple) and isinstance(b, tuple):
				return all(almost(aa, bb) for (aa, bb) in zip(a, b))
			return abs(a - b) < 1

		with Engine(self.prefix) as eng:
			phys = eng.new_character("physical")
			here = phys.new_place((0, 0))
			phys.add_place((1, 1))
			phys.add_place(9)  # test that gridboard can handle this
			this = here.new_thing(2)

			@this.rule(always=True)
			def go(me):
				if me["location"] == (1, 1):
					me["location"] = 9
				elif me["location"] == 9:
					me["location"] = (0, 0)
				else:
					me["location"] = (1, 1)

		app = self.app
		app.starting_dir = self.prefix
		app.build()
		idle_until(
			lambda: hasattr(app, "engine"), 100, "Never got engine proxy"
		)
		assert app.engine.character["physical"].thing[2]["location"] == (0, 0)
		graphboard = app.mainscreen.graphboards["physical"]
		gridboard = app.mainscreen.gridboards["physical"]
		idle_until(
			lambda: graphboard.size != [100, 100],
			100,
			"Never resized graphboard",
		)
		idle_until(
			lambda: gridboard.size != [100, 100],
			100,
			"Never resized gridboard",
		)
		idle_until(
			lambda: (0, 0) in graphboard.spot,
			100,
			"Never made spot for location 0",
		)
		idle_until(
			lambda: 2 in graphboard.pawn, 100, "Never made pawn for thing 2"
		)
		locspot0 = graphboard.spot[0, 0]
		gridspot0 = gridboard.spot[0, 0]
		locspot1 = graphboard.spot[1, 1]
		gridspot1 = gridboard.spot[1, 1]
		graphpawn = graphboard.pawn[2]
		gridpawn = gridboard.pawn[2]
		idle_until(
			lambda: almost(graphpawn.x, locspot0.right),
			100,
			f"Never positioned pawn to 0's right (it's at {graphpawn.x}"
			f", not {locspot0.right})",
		)
		idle_until(
			lambda: almost(graphpawn.y, locspot0.top),
			100,
			"Never positioned pawn to 0's top",
		)
		idle_until(
			lambda: almost(gridpawn.pos, gridspot0.pos),
			100,
			"Never positioned pawn to grid 0, 0",
		)
		app.mainscreen.next_turn()
		idle_until(lambda: not app.edit_locked, 100, "Never unlocked")

		loc = app.engine.character["physical"].thing[2]["location"]

		def relocated_to(dest):
			nonlocal loc
			loc = app.engine.character["physical"].thing[2]["location"]
			return loc == dest

		idle_until(
			partial(relocated_to, (1, 1)),
			1000,
			f"Thing 2 didn't go to location (1, 1); instead, it's at {loc}",
		)
		idle_until(
			lambda: almost(graphpawn.x, locspot1.right),
			100,
			"Never positioned pawn to 1's right "
			f"(pawn is at {graphpawn.x} not {locspot1.right})",
		)
		idle_until(
			lambda: almost(graphpawn.y, locspot1.top),
			100,
			"Never positioned pawn to 1's top "
			f"(it's at {graphpawn.y}, not {locspot1.top})",
		)
		idle_until(
			lambda: almost(gridpawn.pos, gridspot1.pos),
			100,
			"Never positioned pawn to grid 1, 1",
		)
		locspot9 = graphboard.spot[9]
		app.mainscreen.next_turn()
		idle_until(lambda: not app.edit_locked, 100, "Never unlocked")
		loc = app.engine.character["physical"].thing[2]["location"]
		idle_until(
			partial(relocated_to, 9),
			1000,
			f"Thing 2 didn't relocate to 9; it's at {loc}",
		)
		idle_until(
			lambda: 2 not in gridboard.pawn,
			100,
			"pawn never removed from grid",
		)
		idle_until(
			lambda: almost(graphpawn.x, locspot9.right)
			and almost(graphpawn.y, locspot9.top),
			100,
			f"Never positioned pawn to 9's top-right, "
			f"it's at {graphpawn.pos} not {locspot9.right, locspot9.top}",
		)
		app.mainscreen.next_turn()
		idle_until(lambda: not app.edit_locked, 100, "Never unlocked")
		loc = app.engine.character["physical"].thing[2]["location"]
		idle_until(
			partial(relocated_to, (0, 0)),
			1000,
			f"Thing 2 didn't relocate to (0, 0); it's at {loc}",
		)
		idle_until(
			lambda: 2 in gridboard.pawn, 100, "pawn never returned to grid"
		)
		idle_until(
			lambda: almost(graphpawn.x, locspot0.right)
			and almost(graphpawn.y, locspot0.top),
			100,
			f"Never returned to 0's top-right "
			f"(stuck at {graphpawn.pos}, should be "
			f"{locspot0.right, locspot0.top})",
		)
		idle_until(
			lambda: almost(gridpawn.pos, gridspot0.pos),
			100,
			"Never returned to grid 0, 0",
		)



from inspect import getsource
from multiprocessing import freeze_support
from tempfile import mkdtemp

import networkx as nx
from kivy.clock import Clock
from kivy.lang.builder import Builder
from kivy.properties import BooleanProperty, NumericProperty
from networkx import grid_2d_graph

from elide.game import GameApp, GameScreen, GridBoard


def make_grid() -> nx.Graph:
	pass


def game_start(engine: "lisien.Engine") -> None:
	pass



from kivy.base import EventLoop
from kivy.tests import UnitTestTouch

from lisien import Engine

from .util import ELiDEAppTest, idle_until


class StringsEditorTest(ELiDEAppTest):
	def test_strings_editor(self):
		assert "lisien" in self.app.config
		app = self.app
		print("app", id(app))
		self.Window.add_widget(app.build())
		idle_until(
			lambda: hasattr(app, "mainscreen"), 100, "app never got mainscreen"
		)
		idle_until(
			lambda: app.manager.has_screen("timestream"),
			100,
			"timestream never added to manager",
		)
		idle_until(
			lambda: hasattr(app.mainmenu, "configurator"),
			100,
			"DirPicker never got configurator",
		)
		app.mainmenu.configurator.start()  # start with blank world

		def app_has_engine():
			return hasattr(self.app.get_running_app(), "engine")

		idle_until(app_has_engine, 600, "app never got engine")
		idle_until(
			lambda: app.strings.children, 100, "strings never got children"
		)
		idle_until(lambda: app.strings.edbox, 100, "strings never got edbox")
		idle_until(
			lambda: "physical" in app.mainscreen.graphboards,
			100,
			"never got physical in graphboards",
		)
		edbox = app.strings.edbox
		strings_list = edbox.ids.strings_list
		idle_until(
			lambda: strings_list.store, 100, "strings_list never got store"
		)
		strings_ed = edbox.ids.strings_ed
		app.strings.toggle()
		self.advance_frames(10)
		touchy = UnitTestTouch(*strings_ed.ids.stringname.center)
		touchy.touch_down()
		EventLoop.idle()
		touchy.touch_up()
		EventLoop.idle()
		strings_ed.ids.stringname.text = "a string"
		idle_until(
			lambda: strings_ed.name == "a string", 100, "name never set"
		)
		touchier = UnitTestTouch(*strings_ed.ids.string.center)
		touchier.touch_down()
		EventLoop.idle()
		touchier.touch_up()
		self.advance_frames(10)
		strings_ed.ids.string.text = "its value"
		idle_until(
			lambda: strings_ed.source == "its value", 100, "source never set"
		)
		self.advance_frames(10)
		edbox.dismiss()
		app.stop()
		with Engine(self.prefix) as eng:
			assert "a string" in eng.string
			assert eng.string["a string"] == "its value"



from math import sqrt

import networkx as nx
from kivy.tests.common import GraphicUnitTest, UnitTestTouch

from elide.app import ElideApp
from elide.graph.arrow import ArrowPlane
from elide.graph.board import GraphBoard, GraphBoardView
from lisien import Engine
from lisien.facade import CharacterFacade

from ..dummy import Dummy
from ..kivygarden.texturestack import TextureStackPlane
from .util import ELiDEAppTest, idle_until


class FakeEngineProxy:
	def handle(self, *args, **kwargs):
		pass


def pos_near(x0, y0, x1, y1):
	return abs(sqrt(x0**2 + y0**2) - sqrt(x1**2 + y1**2)) < 10


class GraphBoardTest(GraphicUnitTest):
	def test_layout_grid(self):
		spots_wide = 3
		spots_tall = 3
		graph = nx.grid_2d_graph(spots_wide, spots_tall)
		char = CharacterFacade(graph)
		app = ElideApp()
		spotlayout = TextureStackPlane()
		arrowlayout = ArrowPlane()
		board = GraphBoard(
			app=app,
			character=char,
			stack_plane=spotlayout,
			arrow_plane=arrowlayout,
		)
		board.engine = FakeEngineProxy()
		spotlayout.pos = board.pos
		board.bind(pos=spotlayout.setter("pos"))
		spotlayout.size = board.size
		board.bind(size=spotlayout.setter("size"))
		board.add_widget(spotlayout)
		arrowlayout.pos = board.pos
		board.bind(pos=arrowlayout.setter("pos"))
		arrowlayout.size = board.size
		board.bind(size=arrowlayout.setter("size"))
		board.add_widget(arrowlayout)
		board.update()
		boardview = GraphBoardView(board=board)
		self.Window.add_widget(boardview)

		@idle_until(timeout=1000, message="Never finished placing spots")
		def all_spots_placed():
			for x in range(spots_wide):
				for y in range(spots_tall):
					if (x, y) not in board.spot:
						return False
			return True

		# Don't get too picky about the exact proportions of the grid; just make sure the
		# spots are positioned logically with respect to one another
		for name, spot in board.spot.items():
			x, y = name
			if x > 0:
				assert spot.x > board.spot[x - 1, y].x
			if y > 0:
				assert spot.y > board.spot[x, y - 1].y
			if x < spots_wide - 1:
				assert spot.x < board.spot[x + 1, y].x
			if y < spots_tall - 1:
				assert spot.y < board.spot[x, y + 1].y

	def test_select_arrow(self):
		char = CharacterFacade()
		char.add_place(0, _x=0.1, _y=0.1)
		char.add_place(1, _x=0.2, _y=0.1)
		char.add_portal(0, 1)
		app = ElideApp()
		board = GraphBoard(app=app, character=char)
		boardview = GraphBoardView(board=board)
		self.Window.add_widget(boardview)
		idle_until(
			lambda: board.arrow_plane, 100, "GraphBoard never got arrow_plane"
		)
		idle_until(
			lambda: 0 in board.arrow and 1 in board.arrow[0],
			100,
			"GraphBoard never got arrow",
		)
		idle_until(
			lambda: board.arrow_plane.data,
			100,
			"GraphBoard.arrow_plane.data never populated",
		)
		idle_until(
			lambda: board.arrow_plane._bot_left_corner_xs.shape[0] > 0,
			100,
			"GraphBoard.arrow_plane never got bounding boxes",
		)
		ox, oy = board.spot[0].center
		dx, dy = board.spot[1].center
		motion = UnitTestTouch((ox + ((dx - ox) / 2)), dy)
		motion.touch_down()
		motion.touch_up()
		idle_until(
			lambda: app.selection == board.arrow[0][1],
			100,
			"Arrow not selected",
		)

	def test_select_spot(self):
		char = CharacterFacade()
		char.add_place(0, _x=0.1, _y=0.1)
		app = ElideApp()
		board = GraphBoard(app=app, character=char)
		boardview = GraphBoardView(board=board)
		self.Window.add_widget(boardview)
		idle_until(lambda: 0 in board.spot)
		x, y = board.spot[0].center
		motion = UnitTestTouch(x, y)
		motion.touch_down()
		motion.touch_up()
		assert app.selection == board.spot[0]

	def test_select_pawn(self):
		char = CharacterFacade()
		char.add_place(0, _x=0.1, _y=0.1)
		char.add_thing("that", location=0)
		app = ElideApp()
		board = GraphBoard(app=app, character=char)
		boardview = GraphBoardView(board=board)
		self.Window.add_widget(boardview)
		idle_until(lambda: 0 in board.spot and "that" in board.pawn, 100)
		motion = UnitTestTouch(*board.pawn["that"].center)
		motion.touch_down()
		motion.touch_up()
		assert app.selection == board.pawn["that"]

	def test_pawn_drag(self):
		char = CharacterFacade()
		char.add_place(0, _x=0.1, _y=0.1)
		char.add_place(1, _x=0.2, _y=0.1)
		char.add_thing("that", location=0)
		app = ElideApp()
		board = GraphBoard(app=app, character=char)
		boardview = GraphBoardView(board=board)
		self.Window.add_widget(boardview)
		idle_until(
			lambda: 0 in board.spot
			and 1 in board.spot
			and "that" in board.pawn
		)
		that = board.pawn["that"]
		one = board.spot[1]
		touch = UnitTestTouch(*that.center)
		touch.touch_down()
		dist_x = one.center_x - that.center_x
		dist_y = one.center_y - that.center_y
		for i in range(1, 11):
			coef = 1 / i
			x = one.center_x - coef * dist_x
			y = one.center_y - coef * dist_y
			touch.touch_move(x, y)
			self.advance_frames(1)
		touch.touch_move(*one.center)
		self.advance_frames(1)
		touch.touch_up(*one.center)
		idle_until(lambda: that.pos != one.center, 100)
		idle_until(lambda: that.proxy["location"] == 1, 100)

	def test_spot_and_pawn_from_dummy(self):
		char = CharacterFacade()
		app = ElideApp()
		board = GraphBoard(app=app, character=char)
		board._connect_proxy_objects()
		view = GraphBoardView(board=board)
		idle_until(
			lambda: view.plane is not None, 100, "Never made BoardScatterPlane"
		)
		idle_until(
			lambda: board.stack_plane is not None, 100, "Never made StackPlane"
		)
		self.Window.add_widget(view)
		dummy = Dummy(
			name="hello",
			paths=["atlas://rltiles/base/unseen"],
			size=(32, 32),
			pos=(0, 0),
		)
		board.add_widget(dummy)
		idle_until(
			lambda: dummy in board.children, 100, "Dummy didn't get to board"
		)
		dummy_name = dummy.name
		view.spot_from_dummy(dummy)
		idle_until(
			lambda: dummy_name in char.node, 100, "Dummy didn't add place"
		)
		dummy2 = Dummy(
			name="goodbye",
			paths=["atlas://rltiles/base/unseen"],
			pos=dummy.pos,
			size=(32, 32),
		)
		dummy2_name = dummy2.name = "dummy2"
		board.add_widget(dummy2)
		idle_until(
			lambda: dummy2 in board.children,
			100,
			"Dummy 2 didn't get to board",
		)
		idle_until(
			lambda: board.stack_plane.data,
			100,
			"Dummy 2 didn't get into the board's stack_plane",
		)
		view.pawn_from_dummy(dummy2)
		idle_until(
			lambda: dummy2_name in char.thing, 100, "Dummy 2 didn't add thing"
		)
		idle_until(
			lambda: dummy2_name in board.pawn,
			100,
			"Board didn't add pawn for dummy 2",
		)
		spot = board.spot[dummy_name]
		idle_until(
			lambda: board.pawn[dummy2_name].pos == (spot.right, spot.top),
			100,
			"Dummy 2 didn't get to dummy 1",
		)

	def test_pawn_add_new_place(self):
		char = CharacterFacade()
		app = ElideApp()
		board = GraphBoard(app=app, character=char)
		board._connect_proxy_objects()
		boardview = GraphBoardView(board=board)
		self.Window.add_widget(boardview)
		idle_until(lambda: board.stack_plane)
		char.add_place(1, _x=0.2, _y=0.2)
		board.add_spot(1)
		idle_until(lambda: 1 in board.spot, 100, "Didn't make spot")
		char.add_thing("that", location=1)
		idle_until(lambda: "that" in board.pawn, 100, "Didn't make pawn")
		that = board.pawn["that"]
		one = board.spot[1]
		idle_until(
			lambda: pos_near(*getattr(that, "pos", None), one.right, one.top),
			100,
			f"pawn did not locate within 100 ticks. "
			f"Should be at {one.right, one.top}, is at {that.pos}",
		)


class SwitchGraphTest(ELiDEAppTest):
	def test_character_switch_graph(self):
		with Engine(self.prefix) as eng:
			eng.add_character("physical", nx.grid_2d_graph(10, 1))
			eng.add_character("tall", nx.grid_2d_graph(1, 10))
		app = self.app
		self.Window.add_widget(app.build())
		idle_until(
			lambda: hasattr(app, "mainscreen")
			and app.mainscreen.mainview
			and app.mainscreen.statpanel
			and hasattr(app.mainscreen, "gridview")
		)
		idle_until(
			lambda: app.mainscreen.boardview
			in app.mainscreen.mainview.children
		)
		idle_until(lambda: app.mainscreen.boardview.board.children)
		print(
			f"test_character_switch_graph got app {id(app)}, engine proxy {id(app.engine)}"
		)
		assert len(
			set(
				child.x
				for child in app.mainscreen.boardview.board.stack_plane.children
			)
		) == len(app.mainscreen.boardview.board.stack_plane.children)
		app.character_name = "tall"

		def all_x_same():
			if (
				app.mainscreen.boardview.board is None
				or app.mainscreen.boardview.board.stack_plane is None
				or not app.mainscreen.boardview.board.spot
			):
				return False
			first_x = next(
				iter(app.mainscreen.boardview.board.spot.values())
			).x
			return all(
				child.x == first_x
				for child in app.mainscreen.boardview.board.spot.values()
			)

		idle_until(all_x_same, 100, "Never got the new board")
		idle_until(
			lambda: len(
				set(
					child.y
					for child in app.mainscreen.boardview.board.stack_plane.children
				)
			)
			== len(app.mainscreen.boardview.board.stack_plane.children),
			100,
			"New board arranged weird",
		)



from inspect import getsource
from multiprocessing import freeze_support
from tempfile import mkdtemp

from kivy import Logger
from kivy.clock import Clock
from kivy.lang.builder import Builder
from kivy.properties import BooleanProperty, NumericProperty

from elide.game import GameApp, GameScreen, GridBoard
from lisien.exc import OutOfTimelineError


def remove_prefix(s: str, prefix: str):
	"""py3.8 compat"""
	if s.startswith(prefix):
		return s[len(prefix) :]
	return s


def game_start(engine) -> None:
	from random import randint, shuffle

	import networkx as nx

	engine.eternal["nonusage-limit"] = 100
	wide = engine.eternal.setdefault("max-pxcor", 36)
	high = engine.eternal.setdefault("max-pycor", 37)
	initworld = nx.grid_2d_graph(wide, high)
	# world wraps vertically
	for x in range(wide):
		initworld.add_edge((x, high - 1), (x, 0))
	# world wraps horizontally
	for y in range(high):
		initworld.add_edge((wide - 1, y), (0, y))

	locs = list(initworld.nodes.keys())
	shuffle(locs)
	peeps = {}

	for turtle in range(engine.eternal.setdefault("people", 60)):
		initworld.add_node(
			"turtle" + str(turtle),
			awareness=0,
			facing=randint(0, 3),
			location=locs.pop(),
			_image_paths=[
				"atlas://rltiles/base/unseen",
				"atlas://rltiles/body/robe_black",
			],
		)
		peeps["turtle" + str(turtle)] = True

	centers = {}

	for center in range(engine.eternal.setdefault("centers", 20)):
		initworld.add_node(
			"center" + str(center),
			location=locs.pop(),
			nonusage=0,
			_image_paths=["atlas://rltiles/dungeon/dngn_altar_xom"],
		)
		centers["center" + str(center)] = True

	phys = engine.new_character("physical", initworld)
	peep = engine.new_character("people", units={"physical": peeps})
	lit = engine.new_character("literature", units={"physical": centers})

	@peep.unit.rule(always=True)
	def wander(person):
		x, y = person.location.name
		if person["facing"] == 0:
			y += 1
		elif person["facing"] == 1:
			x += 1
		elif person["facing"] == 2:
			y -= 1
		elif person["facing"] == 3:
			x -= 1
		x %= person.engine.eternal["max-pxcor"]
		y %= person.engine.eternal["max-pycor"]
		person["location"] = (x, y)
		person["facing"] = (
			person["facing"]
			+ person.engine.randint(0, 1)
			- person.engine.randint(0, 1)
		)

	@engine.function
	def has_literature(person):
		for contained in person.location.contents():
			if contained.name.startswith("flyer") or contained.name.startswith(
				"center"
			):
				return True
		return False

	@peep.unit.rule
	def learn(person):
		person["awareness"] = (person["awareness"] + 1) % 15
		if person["awareness"] < 5:
			image_paths = [
				"atlas://rltiles/base/unseen",
				"atlas://rltiles/body/robe_black",
			]
		elif person["awareness"] < 10:
			image_paths = [
				"atlas://rltiles/base/unseen",
				"atlas://rltiles/body/robe_green",
			]
		elif person["awareness"] < 15:
			image_paths = [
				"atlas://rltiles/base/unseen",
				"atlas://rltiles/body/robe_white_green",
			]
		else:
			image_paths = [
				"atlas://rltiles/base/unseen",
				"atlas://rltiles/body/robe_green_gold",
			]
		if person["_image_paths"] != image_paths:
			person["_image_paths"] = image_paths
		person["last_learned"] = person.engine.turn
		for contained in person.location.contents():
			if contained.name.startswith("flyer") or contained.name.startswith(
				"center"
			):
				contained["last_read"] = person.engine.turn

	# this would be more pleasant as something like a partial
	@learn.trigger
	def literature_here(person):
		return person.engine.function.has_literature(person)

	@peep.unit.rule
	def unlearn(person):
		person["awareness"] = max((person["awareness"] - 1, 0))

	@unlearn.trigger
	def no_literature_here(person):
		return not person.engine.function.has_literature(person)

	@peep.unit.rule
	def write(person):
		lit_ = person.engine.character["literature"]
		maxnum = 0
		for unit in lit_.units():
			if unit.name.startswith("flyer"):
				maxnum = max((int(remove_prefix(unit.name, "flyer")), maxnum))
		scroll = person.location.new_thing(
			f"flyer{maxnum:02}",
			nonusage=0,
			_image_paths=["atlas://rltiles/scroll/scroll-0"],
		)
		lit_.add_unit(scroll)

	@write.trigger
	def activist(person):
		return person["awareness"] >= 15

	@peep.unit.rule
	def preach(person):
		for that in person.location.contents():
			if that != person and that.name.startswith("turtle"):
				that["awareness"] = (that["awareness"] + 1) % 15

	@preach.trigger
	def well_informed(person):
		return person["awareness"] >= 10

	@lit.unit.rule
	def disappear(ctr):
		ctr.delete()

	@disappear.trigger
	def unused(ctr):
		return (
			ctr.get("last_read", ctr.engine.turn) - ctr.engine.turn
			> ctr.engine.eternal["nonusage-limit"]
		)


class AwarenessGridBoard(GridBoard):
	def on_selection(self, *args):
		if not GameApp.get_running_app().placing_centers or not isinstance(
			self.selection, self.spot_cls
		):
			return
		prox = self.selection.proxy
		for contained in prox.contents():
			if contained.name.startswith("center"):
				return
		name = f"""center{
			max(
				int(remove_prefix(name, "center"))
				for name in self.character.thing.keys()
				if isinstance(name, str) and name.startswith("center")
			)
			+ 1
		}"""
		prox.add_thing(
			name,
			nonusage=0,
			_image_paths=["atlas://rltiles/dungeon/dngn_altar_xom"],
		)


class MainGame(GameScreen):
	def on_parent(self, *args):
		if "game" not in self.ids or "physical" not in self.engine.character:
			if "physical" not in self.engine.character:
				Logger.debug("MainGame.on_parent: no physical character")
			Clock.schedule_once(self.on_parent, 0)
			return
		self.set_up()
		self.ids.game.board = AwarenessGridBoard(
			character=self.engine.character["physical"]
		)
		AwarenessApp.get_running_app().bind(turn=self._get_turn)

	def _get_turn(self, *args):
		app = AwarenessApp.get_running_app()
		if app.turn > app.end_turn:
			app.end_turn = app.turn
		self.ids.timeslider.value = app.turn

	def set_up(self):
		"""Regenerate the whole map"""
		self.engine.turn = 0
		if hasattr(self, "ran_once"):
			self.engine.eternal["people"] = int(self.ids.people.value)
			self.engine.eternal["centers"] = int(self.ids.centers.value)
			self.engine.eternal["nonusage-limit"] = int(
				self.ids.nonusage.value
			)
		else:
			self.pull_values()
		self._push_character()

	def pull_values(self, *_):
		app = GameApp.get_running_app()
		if (
			"people" not in app.engine.eternal
			or "centers" not in app.engine.eternal
			or "nonusage-limit" not in app.engine.eternal
		):
			Logger.debug("eternals not set")
			Clock.schedule_once(self.pull_values, 0)
			return
		app = GameApp.get_running_app()
		self.ids.people.value = app.engine.eternal["people"]
		self.ids.centers.value = app.engine.eternal["centers"]
		self.ids.nonusage.value = app.engine.eternal["nonusage-limit"]
		self.ran_once = True

	def _push_character(self, *args):
		board = self.ids.game.board
		if not board:
			Clock.schedule_once(self._push_character, 0)
			return
		board.character.thing.disconnect(board.update_from_thing)
		phys = AwarenessApp.get_running_app().engine.character["physical"]
		board.character = phys
		board.update()
		phys.thing.connect(board.update_from_thing)


class AwarenessApp(GameApp):
	play = BooleanProperty(False)
	placing_centers = BooleanProperty(False)
	end_turn = NumericProperty(0)
	inspector = True

	def on_play(self, *args):
		if self.play:
			Clock.schedule_interval(self.next_turn, self.turn_length)
			self._scheduled_next_turn = True
		elif hasattr(self, "_scheduled_next_turn"):
			Clock.unschedule(self.next_turn)
			del self._scheduled_next_turn

	def on_turn(self, *args):
		turn = int(self.turn)
		if turn != self.engine.turn:
			try:
				self.engine.turn = turn
			except OutOfTimelineError:
				Logger.warning("Can't go back to turn", turn)

	def on_start(self):
		if not hasattr(self, "_initialized"):
			self._initialized = True
			self.engine.game_init()


kv = """
# kv_start
<ScreenManager>:
	MainGame:
		name: 'play'
<MainGame>:
	BoxLayout:
		orientation: 'vertical'
		BoxLayout:
			orientation: 'horizontal'
			BoxLayout:
				id: sidebar
				orientation: 'vertical'
				width: 333
				size_hint_x: None
				BoxLayout:
					id: controls
					orientation: 'vertical'
					size_hint_y: 0.3
					BoxLayout:
						id: toprow
						Button:
							id: setup
							text: 'setup'
							on_release: root.set_up()
						Slider:
							id: people
							min: 0
							max: 300
							step: 1
							Label:
								text: 'people'
								x: people.x
								y: people.y
								size: self.texture_size
								size_hint: None, None
							Label:
								text: str(int(people.value))
								x: people.right - self.texture_size[0]
								y: people.y
								size: self.texture_size
								size_hint: None, None
						ToggleButton:
							text: 'place centers'
							on_release: app.placing_centers = self.state == 'down'
					BoxLayout:
						id: midrow	
						ToggleButton:
							id: play
							text: 'Go'
							on_state: app.play = self.state == 'down'
						Slider:
							id: centers
							min: 0
							max: 100
							step: 1
							Label:
								text: 'centers'
								size: self.texture_size
								pos: self.parent.pos
							Label:
								text: str(int(self.parent.value))
								size: self.texture_size
								center_x: self.parent.center_x
								y: self.parent.y
						Widget:
							id: filler0
					Slider:
						id: nonusage
						min: 0
						max: 500
						step: 1
						Label:
							text: 'non-usage limit'
							size: self.texture_size
							pos: nonusage.pos
						Label:
							text: '{:d} ticks'.format(int(nonusage.value))
							size: self.texture_size
							x: nonusage.right - self.texture_size[0]
							y: nonusage.y
				Widget:
					id: filler
			Widget:
				id: gamebox
				size_hint_x: 0.7
				GridBoardView:
					id: game		
					pos: gamebox.pos
					size: gamebox.size
		Slider:
			id: timeslider
			x: 0
			y: 0
			height: 50
			size_hint_y: None
			min: 0
			max: app.end_turn
			step: 1
			on_value: app.turn = self.value
# kv_end
"""

Builder.load_string(kv)

if __name__ == "__main__":
	freeze_support()
	d = mkdtemp()
	with open(d + "/game_start.py", "w", encoding="utf-8") as outf:
		outf.write(getsource(game_start))
	AwarenessApp(prefix=d).run()
	print("Files are in " + d)



from collections import defaultdict
from functools import partial
from itertools import chain
from time import monotonic

from kivy.clock import Clock
from kivy.lang.builder import Builder
from kivy.logger import Logger
from kivy.properties import (
	BooleanProperty,
	ListProperty,
	NumericProperty,
	ObjectProperty,
	ReferenceListProperty,
)
from kivy.uix.widget import Widget
from kivy.vector import Vector

from elide.boardscatter import BoardScatterPlane

from ..boardview import BoardView
from ..kivygarden.texturestack import Stack, TextureStackPlane


class GridPawn(Stack):
	default_image_paths = ["atlas://rltiles/base.atlas/unseen"]

	@property
	def _stack_plane(self):
		return self.board.pawn_plane


class GridSpot(Stack):
	default_image_paths = ["atlas://rltiles/floor.atlas/floor-stone"]

	@property
	def _stack_plane(self):
		return self.board.spot_plane


class GridBoard(Widget):
	selection = ObjectProperty()
	selection_candidates = ListProperty()
	character = ObjectProperty()
	tile_width = NumericProperty(32)
	tile_height = NumericProperty(32)
	tile_size = ReferenceListProperty(tile_width, tile_height)
	pawn_cls = GridPawn
	spot_cls = GridSpot

	def __init__(self, **kwargs):
		self.pawn = {}
		self.spot = {}
		self.contained = defaultdict(set)
		super().__init__(**kwargs)

	def update(self):
		make_spot = self.make_spot
		make_pawn = self.make_pawn
		self.spot_plane.data = list(
			map(make_spot, self.character.place.values())
		)
		self.pawn_plane.data = list(
			map(make_pawn, self.character.thing.values())
		)
		self.spot_plane.redraw()
		self.pawn_plane.redraw()

	def add_spot(self, placen, *args):
		if placen not in self.character.place:
			raise KeyError(f"No such place for spot: {placen}")
		if placen in self.spot:
			raise KeyError("Already have a Spot for this Place")
		spt = self.make_spot(self.character.place[placen])
		self.spot_plane.add_datum(spt)
		self.spot[placen] = self.spot_cls(board=self, proxy=spt["proxy"])

	def make_spot(self, place):
		placen = place["name"]
		if not isinstance(placen, tuple) or len(placen) != 2:
			raise TypeError(
				"Can only make spot from places with tuple names of length 2",
				placen,
			)
		if (
			not isinstance(placen, tuple)
			or len(placen) != 2
			or not isinstance(placen[0], int)
			or not isinstance(placen[1], int)
		):
			raise TypeError(
				"GridBoard can only display places named with pairs of ints"
			)
		if "_image_paths" in place:
			textures = list(place["_image_paths"])
		else:
			textures = list(self.spot_cls.default_image_paths)
		r = {
			"name": placen,
			"x": int(placen[0] * self.tile_width),
			"y": int(placen[1] * self.tile_height),
			"width": int(self.tile_width),
			"height": int(self.tile_height),
			"textures": textures,
			"proxy": place,
		}
		return r

	def make_pawn(self, thing) -> dict:
		location = self.spot[thing["location"]]
		r = {
			"name": thing["name"],
			"x": int(location.x),
			"y": int(location.y),
			"width": int(self.tile_width),
			"height": int(self.tile_height),
			"location": location,
			"textures": list(
				thing.get("_image_paths", self.pawn_cls.default_image_paths)
			),
			"proxy": thing,
		}
		return r

	def add_pawn(self, thingn, *args):
		if thingn not in self.character.thing:
			raise KeyError(f"No such thing: {thingn}")
		if thingn in self.pawn:
			raise KeyError(f"Already have a pawn for {thingn}")
		thing = self.character.thing[thingn]
		if thing["location"] not in self.spot:
			# The location is not in the grid. That's fine.
			return
		pwn = self.make_pawn(thing)
		self.pawn[thingn] = self.pawn_cls(board=self, proxy=pwn["proxy"])
		location = pwn["location"]
		self.contained[location].add(thingn)
		self.pawn_plane.add_datum(pwn)

	def _trigger_add_pawn(self, thingn):
		part = partial(self.add_pawn, thingn)
		Clock.unschedule(part)
		Clock.schedule_once(part, 0)

	def on_parent(self, *args):
		if self.character is None:
			Clock.schedule_once(self.on_parent, 0)
			return
		if self.character.engine.closed:
			return
		Logger.debug("GridBoard: on_parent start")
		start_ts = monotonic()
		if not hasattr(self, "pawn_plane"):
			self.pawn_plane = TextureStackPlane(pos=self.pos, size=self.size)
			self.spot_plane = TextureStackPlane(pos=self.pos, size=self.size)
			self.bind(
				pos=self.pawn_plane.setter("pos"),
				size=self.pawn_plane.setter("size"),
			)
			self.bind(
				pos=self.spot_plane.setter("pos"),
				size=self.spot_plane.setter("size"),
			)
			self.add_widget(self.spot_plane)
			self.add_widget(self.pawn_plane)
		spot_data = list(
			map(
				self.make_spot,
				filter(
					lambda spot: isinstance(spot["name"], tuple)
					and len(spot["name"]) == 2,
					self.character.place.values(),
				),
			)
		)
		if not spot_data:
			if hasattr(self.spot_plane, "_redraw_bind_uid"):
				self.spot_plane.unbind_uid(
					"data", self.spot_plane._redraw_bind_data_uid
				)
				del self.spot_plane._redraw_bind_data_uid
			if hasattr(self.pawn_plane, "_redraw_bind_uid"):
				self.pawn_plane.unbind_uid(
					"data", self.pawn_plane._redraw_bind_data_uid
				)
				del self.pawn_plane._redraw_bind_data_uid
			self.spot_plane.data = self.pawn_plane.data = []
			self.spot_plane.redraw()
			self.pawn_plane.redraw()
			self.spot_plane._redraw_bind_data_uid = self.spot_plane.fbind(
				"data", self.spot_plane._trigger_redraw
			)
			self.pawn_plane._redraw_bind_data_uid = self.pawn_plane.fbind(
				"data", self.pawn_plane._trigger_redraw
			)
			self.character.thing.connect(self.update_from_thing)
			self.character.place.connect(self.update_from_place)
			return
		for spt in spot_data:
			self.spot[spt["name"]] = self.spot_cls(
				board=self, proxy=spt["proxy"]
			)
		self.spot_plane.unbind_uid(
			"data", self.spot_plane._redraw_bind_data_uid
		)
		self.spot_plane.data = spot_data
		self.spot_plane.redraw()
		self.spot_plane._redraw_bind_data_uid = self.spot_plane.fbind(
			"data", self.spot_plane._trigger_redraw
		)
		wide = max(datum["x"] for datum in spot_data) + self.tile_width
		high = max(datum["y"] for datum in spot_data) + self.tile_width
		self.size = self.spot_plane.size = self.pawn_plane.size = wide, high
		pawn_data = list(
			map(
				self.make_pawn,
				filter(
					lambda thing: thing["location"] in self.spot,
					self.character.thing.values(),
				),
			)
		)
		for pwn in pawn_data:
			self.pawn[pwn["name"]] = self.pawn_cls(
				board=self, proxy=pwn["proxy"]
			)
		self.pawn_plane.data = pawn_data
		self.character.thing.connect(self.update_from_thing)
		self.character.place.connect(self.update_from_place)
		Logger.debug(
			f"GridBoard: on_parent end, took {monotonic() - start_ts:,.2f}"
			f" seconds"
		)

	def rm_spot(self, name):
		spot = self.spot.pop(name)
		if spot in self.selection_candidates:
			self.selection_candidates.remove(spot)
		for thing in spot.proxy.contents():
			self.rm_pawn(thing.name)
		self.spot_plane.remove(name)

	def rm_pawn(self, name):
		pwn = self.pawn.pop(name)
		if pwn in self.selection_candidates:
			self.selection_candidates.remove(pwn)
		self.pawn_plane.remove(name)

	def update_from_thing(self, thing, key, value):
		if thing and not (key is None and not value):
			if thing.name not in self.pawn:
				self.add_pawn(thing.name)
			elif key == "location":
				if value in self.spot:
					loc = self.spot[value]
					pwn = self.pawn[thing.name]
					pwn.pos = loc.pos
				elif thing.name in self.pawn:
					self.rm_pawn(thing.name)
			elif key == "_image_paths":
				self.pawn[thing.name].paths = value
		else:
			if thing.name in self.pawn:
				self.rm_pawn(thing.name)

	def update_from_place(self, place, key, value):
		if place and not (key is None and not value):
			if key == "_image_paths":
				self.spot[place.name].paths = value
		else:
			if place.name in self.spot:
				self.rm_spot(place.name)

	def on_touch_down(self, touch):
		self.selection_candidates.extend(
			chain(
				map(
					self.pawn.get,
					self.pawn_plane.iter_collided_keys(*touch.pos),
				),
				map(
					self.spot.get,
					self.spot_plane.iter_collided_keys(*touch.pos),
				),
			)
		)
		return super().on_touch_down(touch)

	def on_touch_up(self, touch):
		touched = {
			candidate
			for candidate in self.selection_candidates
			if candidate.collide_point(*touch.pos)
		}
		if not touched:
			self.selection_candidates = []
			return super().on_touch_up(touch)
		if len(touched) == 1:
			self.selection = touched.pop()
			self.selection_candidates = []
			return super().on_touch_up(touch)
		pawns_touched = {
			node for node in touched if isinstance(node, self.pawn_cls)
		}
		if len(pawns_touched) == 1:
			self.selection = pawns_touched.pop()
			self.selection_candidates = []
			return super().on_touch_up(touch)
		elif pawns_touched:
			# TODO: Repeatedly touching a spot with multiple pawns on it
			# 		should cycle through the pawns, and then finally the spot.
			self.selection = pawns_touched.pop()
			self.selection_candidates = []
			return super().on_touch_up(touch)
		spots_touched = touched - pawns_touched
		if len(spots_touched) == 1:
			self.selection = spots_touched.pop()
			self.selection_candidates = []
			return super().on_touch_up(touch)
		assert not spots_touched, (
			"How do you have overlapping spots on a GridBoard??"
		)
		self.selection_candidates = []
		return super().on_touch_up(touch)


class GridBoardScatterPlane(BoardScatterPlane):
	selection_candidates = ListProperty([])
	selection = ObjectProperty(allownone=True)
	keep_selection = BooleanProperty(False)
	board = ObjectProperty()

	def spot_from_dummy(self, dummy):
		raise NotImplementedError("oop")

	def pawn_from_dummy(self, dummy):
		dummy_center = self.to_local(*dummy.center)
		candidates = list(
			self.board.spot_plane.iter_collided_keys(*dummy_center)
		)
		if not candidates:
			return
		whereat_d = self.board.spot[candidates.pop()]
		half_wide = self.board.tile_width / 2
		half_high = self.board.tile_height / 2
		if candidates:
			whereat_center = whereat_d.x + half_wide, whereat_d.y + half_high
			dist = Vector(*whereat_center).distance(dummy_center)
			while candidates:
				thereat_d = self.board.spot[candidates.pop()]
				thereat_center = (
					thereat_d.x + half_wide,
					thereat_d.y + half_high,
				)
				thereto = Vector(*thereat_center).distance(dummy_center)
				if thereto < dist:
					whereat_d, dist = thereat_d, thereto
		self.board.pawn_plane.add_datum(
			self.board.make_pawn(
				self.board.character.new_thing(
					dummy.name, whereat_d.name, _image_paths=list(dummy.paths)
				)
			)
		)
		dummy.num += 1

	def on_board(self, *args):
		self.clear_widgets()
		self.add_widget(self.board)


class GridBoardView(BoardView):
	pass


kv = """
<GridBoard>:
	app: app
	size_hint: None, None
<GridBoardView>:
	plane: boardplane
	GridBoardScatterPlane:
		id: boardplane
		board: root.board
		scale_min: root.scale_min
		scale_max: root.scale_max
		pos: root.pos
		size: root.size
"""
kv_loaded = False
if not kv_loaded:
	Builder.load_string(kv)
	kv_loaded = True



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from kivy.properties import ListProperty
from kivy.uix.modalview import ModalView

from .util import load_string_once


class KeywordListModal(ModalView):
	data = ListProperty([])


Builder.load_string("""
<KeywordListModal>:
	size_hint_x: 0.6
	BoxLayout:
		orientation: 'vertical'
		StatListView:
			data: root.data
		BoxLayout:
			Button:
				text: 'Cancel'
			Button:
				text: 'Done'
""")



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from kivy.clock import Clock, triggered
from kivy.logger import Logger
from kivy.properties import (
	ListProperty,
	NumericProperty,
	ObjectProperty,
	StringProperty,
)
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.label import Label
from kivy.uix.screenmanager import Screen
from kivy.uix.scrollview import ScrollView

from .kivygarden.texturestack import ImageStack
from .pallet import Pallet, PalletBox
from .util import load_string_once

# TODO: let the user import their own sprite art


def trigger(func):
	return triggered()(func)


class SpriteSelector(BoxLayout):
	prefix = StringProperty()
	pallets = ListProperty()
	imgpaths = ListProperty([])
	default_imgpaths = ListProperty()
	preview = ObjectProperty()

	def on_prefix(self, *_):
		if "textbox" not in self.ids:
			Clock.schedule_once(self.on_prefix, 0)
			return
		self.ids.textbox.text = self.prefix

	def on_imgpaths(self, *_):
		if not self.preview:
			Logger.debug("SpriteSelector: no preview")
			Clock.schedule_once(self.on_imgpaths, 0)
			return
		if hasattr(self, "_imgstack"):
			self._imgstack.paths = self.imgpaths
		else:
			self._imgstack = ImageStack(paths=self.imgpaths)
			self._imgstack.bind(
				pos=self._position_imgstack, size=self._position_imgstack
			)
			self.preview.add_widget(self._imgstack)

	@trigger
	def _position_imgstack(self, *_):
		self._imgstack.x = self.preview.center_x - self._imgstack.height / 2
		self._imgstack.y = self.preview.center_y - self._imgstack.width / 2

	def on_pallets(self, *_):
		for pallet in self.pallets:
			pallet.fbind("selection", self._upd_imgpaths)

	def _upd_imgpaths(self, *_):
		imgpaths = []
		for pallet in self.pallets:
			if pallet.selection:
				for selected in pallet.selection:
					imgpaths.append(
						"atlas://{}/{}".format(pallet.filename, selected.text)
					)
		self.imgpaths = imgpaths if imgpaths else self.default_imgpaths


class SpriteBuilder(ScrollView):
	prefix = StringProperty()
	imgpaths = ListProperty()
	default_imgpaths = ListProperty()
	data = ListProperty()
	labels = ListProperty()
	pallets = ListProperty()

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.bind(data=self._trigger_update)

	def update(self, *args):
		if self.data is None:
			return
		if not self.canvas:
			Clock.schedule_once(self.update, 0)
			return
		if not hasattr(self, "_palbox"):
			self._palbox = PalletBox(orientation="vertical", size_hint_y=None)
			self.add_widget(self._palbox)
		else:
			self._palbox.clear_widgets()
		if hasattr(self._palbox, "_bound_width"):
			for uid in self._palbox._bound_width:
				self._palbox.unbind_uid("width", uid)
			del self._palbox._bound_width
		self.labels = []
		for pallet in self.pallets:
			if hasattr(pallet, "_bound_minimum_height"):
				pallet.unbind_uid(
					"minimum_height", pallet._bound_minimum_height
				)
				del pallet._bound_minimum_height
			if hasattr(pallet, "_bound_height"):
				pallet.unbind_uid("height", pallet._bound_height)
				del pallet._bound_height
		self.pallets = []
		for text, filename in self.data:
			label = Label(text=text, size_hint=(None, None), halign="center")
			label.texture_update()
			label.height = label.texture.height
			label.width = self._palbox.width
			pallet = Pallet(filename=filename, size_hint=(None, None))
			pallet.width = self._palbox.width
			self._palbox._bound_width = [
				self._palbox.fbind("width", label.setter("width")),
				self._palbox.fbind("width", pallet.setter("width")),
			]
			pallet.height = pallet.minimum_height
			pallet._bound_minimum_height = (
				pallet.fbind("minimum_height", pallet.setter("height")),
			)
			pallet._bound_height = pallet.fbind(
				"height", self._trigger_reheight
			)
			self.labels.append(label)
			self.pallets.append(pallet)
		n = len(self.labels)
		assert n == len(self.pallets)
		for i in range(0, n):
			self._palbox.add_widget(self.labels[i])
			self._palbox.add_widget(self.pallets[i])

	_trigger_update = trigger(update)

	def reheight(self, *args):
		self._palbox.height = sum(
			wid.height for wid in self.labels + self.pallets
		)

	_trigger_reheight = trigger(reheight)


class SpriteDialog(BoxLayout):
	toggle = ObjectProperty()
	prefix = StringProperty()
	imgpaths = ListProperty()
	default_imgpaths = ListProperty()
	data = ListProperty()
	pallet_box_height = NumericProperty()

	def pressed(self):
		self.prefix = self.ids.selector.prefix
		self.imgpaths = self.ids.selector.imgpaths
		self.toggle()


class PawnConfigDialog(SpriteDialog):
	pass


class SpotConfigDialog(SpriteDialog):
	pass


class PawnConfigScreen(Screen):
	toggle = ObjectProperty()
	data = ListProperty()
	imgpaths = ListProperty()


class SpotConfigScreen(Screen):
	toggle = ObjectProperty()
	data = ListProperty()
	imgpaths = ListProperty()


load_string_once("""
<SpriteDialog>:
	orientation: 'vertical'
	SpriteBuilder:
		id: builder
		prefix: root.prefix
		default_imgpaths: root.default_imgpaths
		imgpaths: root.imgpaths
		data: root.data
	SpriteSelector:
		id: selector
		textbox: textbox
		size_hint_y: 0.1
		prefix: root.prefix
		default_imgpaths: root.default_imgpaths
		imgpaths: root.imgpaths
		pallets: builder.pallets
		preview: preview
		TextInput:
			id: textbox
			multiline: False
			write_tab: False
			hint_text: 'Enter name prefix'
		Widget:
			id: preview
			canvas:
				Color:
					rgba: 1, 1, 1, 1
		Button:
			text: 'OK'
			on_release: root.pressed()
<PawnConfigScreen>:
	name: 'pawncfg'
	imgpaths: dialog.imgpaths
	PawnConfigDialog:
		id: dialog
		toggle: root.toggle
		default_imgpaths: ['atlas://rltiles/base/unseen']
		data: root.data
<SpotConfigScreen>:
	name: 'spotcfg'
	imgpaths: dialog.imgpaths
	SpotConfigDialog:
		id: dialog
		toggle: root.toggle
		default_imgpaths: ['atlas://rltiles/floor/floor-stone']
		data: root.data
""")



from abc import abstractmethod
from functools import partial

from kivy.tests.common import UnitTestTouch

from lisien import Engine
from lisien.examples import kobold, polygons

from ..card import Card, Foundation
from .util import ELiDEAppTest, idle_until


def builder_foundation(builder):
	for child in builder.children:
		if isinstance(child, Foundation):
			return True
	return False


class RuleBuilderTest(ELiDEAppTest):
	@abstractmethod
	def install(self, engine: Engine):
		raise NotImplementedError()

	@abstractmethod
	def get_selection(self):
		raise NotImplementedError()

	def setUp(self):
		super(RuleBuilderTest, self).setUp()
		with Engine(self.prefix) as eng:
			self.install(eng)
		app = self.app
		mgr = app.build()
		self.Window.add_widget(mgr)
		screen = app.mainscreen
		idle_until(lambda: app.rules.rulesview, 100, "Never made rules view")
		idle_until(
			lambda: "physical" in screen.graphboards,
			100,
			"Never made physical board",
		)
		self.board = screen.graphboards["physical"]
		idle_until(
			lambda: "kobold" in self.board.pawn,
			100,
			"never got the pawn for the kobold",
		)
		app.selection = self.get_selection()
		screen.charmenu.charmenu.toggle_rules()
		rules = app.rules
		rules_box = rules.children[0]
		idle_until(
			lambda: "ruleslist" in rules_box.ids, 100, "Never made rules list"
		)
		self.rules_list = rules_list = rules_box.ids.ruleslist
		self.rules_view = rules_box.ids.rulesview
		idle_until(
			lambda: rules_list.children[0].children,
			100,
			"Never filled rules list",
		)


class TestRuleBuilderKobold(RuleBuilderTest):
	def install(self, engine: Engine):
		kobold.inittest(engine)

	def get_selection(self):
		return self.board.pawn["kobold"]

	def test_rule_builder_display_trigger(self):
		rules_list = self.rules_list
		rules_view = self.rules_view
		idle_until(
			lambda: "shrubsprint"
			in {rulebut.text for rulebut in rules_list.children[0].children},
			100,
			"Never made shrubsprint button",
		)
		for rulebut in rules_list.children[0].children:
			if rulebut.text == "shrubsprint":
				rulebut.state = "down"
				break
		idle_until(lambda: rules_view.children)
		idle_until(
			lambda: hasattr(rules_view, "_trigger_tab"),
			100,
			"Never made trigger tab",
		)
		builder = rules_view._trigger_builder
		idle_until(
			lambda: [
				child for child in builder.children if isinstance(child, Card)
			],
			100,
			"Never filled trigger builder",
		)
		card_names = {
			card.headline_text
			for card in builder.children
			if isinstance(card, Card)
		}
		assert card_names == {
			"standing_still",
			"aware",
			"uncovered",
			"sametile",
			"kobold_alive",
		}

	def test_rule_builder_remove_trigger(self):
		rules_list = self.rules_list
		rules_view = self.rules_view
		idle_until(
			lambda: "shrubsprint"
			in {rulebut.text for rulebut in rules_list.children[0].children},
			100,
			"Never made shrubsprint button",
		)
		for rulebut in rules_list.children[0].children:
			if rulebut.text == "shrubsprint":
				rulebut.state = "down"
				break
		idle_until(lambda: rules_view.children)
		idle_until(
			lambda: hasattr(rules_view, "_trigger_tab"),
			100,
			"Never made trigger tab",
		)
		builder = rules_view._trigger_builder
		idle_until(
			lambda: [
				child for child in builder.children if isinstance(child, Card)
			],
			100,
			"Never filled trigger builder",
		)

		uncovered = None

		def have_uncovered():
			nonlocal uncovered
			for card in builder.children:
				if not isinstance(card, Card):
					continue
				if card.headline_text == "uncovered":
					uncovered = card
					return True
			return False

		idle_until(have_uncovered, 100, "Never got 'uncovered' card")

		right_foundation = None

		def have_right_foundation():
			nonlocal right_foundation
			for foundation in builder.children:
				if not isinstance(foundation, Foundation):
					continue
				if foundation.x > uncovered.right:
					right_foundation = foundation
					return True
			return False

		idle_until(have_right_foundation, 100, "Never built right foundation")

		assert uncovered is not None
		assert right_foundation is not None

		def uncovered_is_flush_with_right_foundation():
			for card in builder.children:
				if not isinstance(card, Card):
					continue
				if card.headline_text == "uncovered":
					breakcover = card
					right_foundation = None
					for foundation in builder.children:
						if isinstance(foundation, Card):
							continue
						if (
							right_foundation is None
							or foundation.x > right_foundation.x
						):
							right_foundation = foundation
					assert right_foundation is not None, "No foundations??"
					return breakcover.x == right_foundation.x
			return False

		card = uncovered
		foundation = right_foundation
		mov = UnitTestTouch(*card.center)
		mov.touch_down()
		dist_x = foundation.center_x - card.center_x
		dist_y = foundation.y - card.center_y
		for i in range(1, 11):
			coef = 1 / i
			x = foundation.center_x - coef * dist_x
			y = foundation.y - coef * dist_y
			mov.touch_move(x, y)
			self.advance_frames(1)
		mov.touch_up(foundation.center_x, foundation.y)
		idle_until(
			partial(builder_foundation, builder),
			100,
			"didn't replace foundations",
		)
		idle_until(
			uncovered_is_flush_with_right_foundation, 100, "card didn't move"
		)
		idle_until(
			lambda: not any(
				func.name == "breakcover"
				for func in self.app.engine.rule["shrubsprint"].triggers
			),
			100,
			"breakcover never removed from rulebook",
		)

	def test_rule_builder_add_trigger(self):
		rules_list = self.rules_list
		rules_view = self.rules_view
		idle_until(
			lambda: "shrubsprint"
			in {rulebut.text for rulebut in rules_list.children[0].children},
			100,
			"Never made shrubsprint button",
		)
		for rulebut in rules_list.children[0].children:
			if rulebut.text == "shrubsprint":
				rulebut.state = "down"
				break
		idle_until(lambda: rules_view.children)
		idle_until(
			lambda: hasattr(rules_view, "_trigger_tab"),
			100,
			"Never made trigger tab",
		)
		builder = rules_view._trigger_builder
		idle_until(
			lambda: [
				child for child in builder.children if isinstance(child, Card)
			],
			100,
			"Never filled trigger builder",
		)
		idle_until(
			lambda: [child.x for child in builder.children if child.x > 0],
			100,
			"Never positioned trigger builder's children",
		)
		aware = None
		for card in builder.children:
			if isinstance(card, Foundation):
				continue
			assert isinstance(card, Card)
			if card.headline_text == "aware":
				aware = card
				break
		assert aware is not None, "Didn't get 'aware' card"
		uncovered = None
		for card in builder.children:
			if isinstance(card, Foundation):
				continue
			assert isinstance(card, Card)
			if card.headline_text == "uncovered":
				uncovered = card
				break
		assert uncovered is not None, "Didn't get 'uncovered' card"
		start_x = aware.center_x
		start_y = aware.top - 10
		assert aware.collide_point(start_x, start_y)
		mov = UnitTestTouch(start_x, start_y)
		mov.touch_down()
		dist_x = start_x - uncovered.center_x
		dist_y = start_y - uncovered.center_y
		decr_x = dist_x / 10
		decr_y = dist_y / 10
		x = start_x
		y = start_y
		for i in range(1, 11):
			x -= decr_x
			y -= decr_y
			mov.touch_move(x, y)
			self.advance_frames(1)
		mov.touch_up(*uncovered.center)
		idle_until(
			lambda: abs(aware.x - uncovered.x) < 2,
			100,
			"aware didn't move to its new place",
		)
		idle_until(
			lambda: any(
				func.name == "aware"
				for func in self.app.engine.rule["shrubsprint"].triggers
			),
			100,
			"aware never added to rulebook",
		)


class TestCharRuleBuilder(ELiDEAppTest):
	def setUp(self):
		with Engine(self.prefix) as eng:
			polygons.install(eng)
			assert list(
				eng.character["triangle"].unit.rule["relocate"].triggers
			) == [
				eng.trigger.similar_neighbors,
				eng.trigger.dissimilar_neighbors,
			]
		super(TestCharRuleBuilder, self).setUp()
		app = self.app
		mgr = app.build()
		self.Window.add_widget(mgr)
		idle_until(
			lambda: hasattr(app, "engine"), 100, "App never made engine"
		)
		idle_until(
			lambda: "triangle" in app.engine.character,
			100,
			"Engine proxy never made triangle character proxy",
		)
		app.select_character(app.engine.character["triangle"])
		idle_until(
			lambda: app.character_name == "triangle",
			100,
			"Never changed character",
		)
		app.mainscreen.charmenu.charmenu.toggle_rules()
		idle_until(
			lambda: getattr(app.charrules, "_finalized", False),
			100,
			"Never finalized",
		)

	def test_char_rule_builder_remove_unit_trigger(self):
		app = self.app
		idle_until(
			lambda: getattr(app.charrules, "_finalized", False),
			100,
			"Never finalized charrules",
		)
		tabitem = app.charrules._unit_tab
		idle_until(lambda: tabitem.content, 100, "unit tab never got content")
		tabitem.on_press()
		self.advance_frames(1)
		tabitem.on_release()
		idle_until(
			lambda: app.charrules._tabs.current_tab == tabitem,
			100,
			"Never switched tab",
		)
		rules_box = app.charrules._unit_box
		idle_until(lambda: rules_box.parent, 100, "unit box never got parent")
		idle_until(
			lambda: getattr(rules_box.rulesview, "_finalized", False),
			100,
			"Never finalized unit rules view",
		)
		idle_until(
			lambda: rules_box.children, 100, "_unit_box never got children"
		)
		idle_until(
			lambda: rules_box.rulesview.children,
			100,
			"Never filled rules view",
		)
		rules_list = rules_box.ruleslist
		idle_until(
			lambda: rules_list.children[0].children,
			1000,
			"Never filled rules list",
		)
		idle_until(
			lambda: "relocate"
			in {rulebut.text for rulebut in rules_list.children[0].children},
			1000,
			"Never made relocate button",
		)
		for rulebut in rules_list.children[0].children:
			if rulebut.text == "relocate":
				rulebut.state = "down"
				break
		builder = rules_box.rulesview._trigger_builder
		assert (
			rules_box.rulesview._tabs.current_tab
			== rules_box.rulesview._trigger_tab
		)
		idle_until(
			lambda: builder.children,
			1000,
			"trigger builder never got children",
		)
		idle_until(
			partial(builder_foundation, builder),
			100,
			"Never filled trigger builder",
		)
		idle_until(
			lambda: builder.parent, 1000, "trigger builder never got parent"
		)
		card_names = {
			card.headline_text
			for card in builder.children
			if isinstance(card, Card)
		}
		assert card_names == {
			"similar_neighbors",
			"dissimilar_neighbors",
		}
		for card in builder.children:
			if not isinstance(card, Card):
				continue
			if card.headline_text == "similar_neighbors":
				break
		else:
			assert False, "Didn't get similar_neighbors"
		startx = card.center_x
		starty = card.top - 1
		assert card.collide_point(startx, starty), "card didn't collide itself"
		for cardother in builder.children:
			if not isinstance(cardother, Card) or cardother == card:
				continue
			assert not cardother.collide_point(startx, starty), (
				"other card will grab the touch"
			)
		touch = UnitTestTouch(startx, starty)
		for target in builder.children:
			if isinstance(target, Card):
				continue
			if target.x > card.right:
				break
		else:
			assert False, "Didn't get target foundation"
		targx, targy = target.center
		distx = targx - startx
		disty = targy - starty
		x, y = startx, starty
		touch.touch_down()
		self.advance_frames(1)
		for i in range(1, 11):
			x += distx / 10
			y += disty / 10
			touch.touch_move(x, y)
			self.advance_frames(1)
		touch.touch_up()
		self.advance_frames(5)
		rules_box.ids.closebut.on_release()
		idle_until(
			lambda: all(
				card.headline_text != "similar_neighbors"
				for card in builder.decks[0]
			),
			100,
			"similar_neighbors still in used pile",
		)
		idle_until(
			lambda: not any(
				trig.name == "similar_neighbors"
				for trig in app.charrules.character.unit.rulebook[0].triggers
			),
			100,
			"similar_neighbors still in proxy triggers list",
		)
		app.stop()
		with Engine(self.prefix) as eng:
			assert list(
				eng.character["triangle"].unit.rule["relocate"].triggers
			) == [eng.trigger.dissimilar_neighbors]
